/*

*/
import scene.*;
import scene.models.*;
import pipeline.*;
import framebuffer.*;

import java.io.File;
import java.awt.Color;
import java.awt.event.*;
import java.util.ArrayList;

/**

*/
public class InteractiveModels_v11
{
   private static boolean render_z_buffer = false;

   private static int perspProjMatrix1Index = 0;

   private static double eyeX = 0.0;
   private static double eyeY = 0.0;
   private static double eyeZ = 2.0;

   private static double near   =  1.0;
   private static double left   = -1.0;
   private static double right  =  1.0;
   private static double bottom = -1.0;
   private static double top    =  1.0;

   private static double xTranslation = 0.0;
   private static double yTranslation = 0.0;
   private static double zTranslation = 0.0;
   private static double xRotation = 0.0;
   private static double yRotation = 0.0;
   private static double zRotation = 0.0;
   private static double scale = 1.0;

   private static ArrayList<Model> modelArray = new ArrayList<Model>();

   public static void main(String[] args)
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the digit keys '0' - '9' to choose a model.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the model along the x, y, z axes.");
      System.out.println("Use the u/U, v/V, w/W, keys to rotate the model around the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the model.");
      System.out.println("Use the 'c' key to change the solid model colors.");
      System.out.println("Use the 'e' key to change the random edge colors.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the '/' key to cycle through the perspective projection matrices.");
      System.out.println("Use the ' ' key to toggle between the image buffer and z-buffer.");
      System.out.println("Use the n/N keys to move the camera's near plane.");
      System.out.println("Use the o/O keys to change the size of the camera's view rectangle .");
      System.out.println("Use the arrow keys to move the camera location up/down and left/right.");

      // Create several Model objects.
      modelArray.add( new ObjSimpleModel(new File("assets/apple.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/cow.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/galleon.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/teapot.obj")) );
//      modelArray.add( new ObjSimpleModel(new File("assets/cessna.obj")) );
//      modelArray.add( new ObjSimpleModel(new File("assets/horse.obj")) );
//      modelArray.add( new SphereModel(30, 30, 1) );
      modelArray.add( new CylinderModel(20, 20, 0.5, 1) );
      modelArray.add( new TorusModel(25, 25, 0.75, 0.25) );
      modelArray.add( new ObjSimpleModel(new File("assets/small_rhombicosidodecahedron.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/great_rhombicosidodecahedron.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/cessna.obj")) );


      // Push the last model off to the side.
      modelArray.get(modelArray.size() - 1).modelTranslate(3, 0, 0);

      // Give the models random colors.
      for (Model m : modelArray)
      {
         m.setColorRandom();
      }


      // Create the Scene object that we shall render.
      final Scene scene = new Scene();

      // Add three models to the Scene.
      scene.addModel( modelArray.get(1),
                      modelArray.get(modelArray.size() - 2),
                      modelArray.get(modelArray.size() - 1) );


      // Create a Camera object.
      Camera camera = new Camera();

      // Set up the camera's view volume.
      camera.projPerspective(left, right, bottom, top, near);
/*
      // An alternative way to set up the camera's view frustum.
      double fov    = 90.0;
      double aspect = 1.0;
      camera.projPerspective(fov, aspect, near);
*/
      // Set up the camera's location and orientation.
      double centerX = 0, centerY = 0, centerZ = 0;
      double     upX = 0,     upY = 1,     upZ = 0;
      camera.viewLookAt(eyeX,    eyeY,    eyeZ,
                        centerX, centerY, centerZ,
                        upX,     upY,     upZ);

      // Add the Camera to the Scene.
      scene.setCamera(camera);


      // Create a InteractiveFrame and a FrameBuffer to render our scene into.
      int width  = 512;
      int height = 512;

      @SuppressWarnings("serial")
      final InteractiveFrame app = new InteractiveFrame("Renderer 11", width, height)
      {
         // Implement part of the KeyListener interface.
         @Override public void keyPressed(KeyEvent e)
         {
            //System.out.println( e );
            int keyCode = e.getKeyCode();
            if (KeyEvent.VK_UP == keyCode)
            {
               eyeY += 0.1;
            }
            else if (KeyEvent.VK_DOWN == keyCode)
            {
               eyeY -= 0.1;
            }
            else if (KeyEvent.VK_LEFT == keyCode)
            {
               eyeX -= 0.1;
            }
            else if (KeyEvent.VK_RIGHT == keyCode)
            {
               eyeX += 0.1;
            }

            // Set up the camera's location and orientation.
            scene.camera.view2Identity();
            double centerX = 0, centerY = 0, centerZ = 0;
            double upX = 0,     upY = 1,     upZ = 0;
            scene.camera.viewLookAt(eyeX,    eyeY,    eyeZ,
                                    centerX, centerY, centerZ,
                                    upX,     upY,     upZ);

            // Render again.
            this.getFrameBuffer().clearFB(Color.black);
            Pipeline.render(scene, this.getFrameBuffer());
            if ( render_z_buffer )
            {
               FrameBuffer zBuf = this.getFrameBuffer().convertZB2FB();
               this.getFrameBuffer().setViewport(0, 0, zBuf);
            }
            repaint();
         }

         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
            //System.out.println( e );
            char c = e.getKeyChar();
            if ('d' == c)
            {
             //Pipeline.debug = ! Pipeline.debug;
               Pipeline.debug_visible = ! Pipeline.debug_visible;
             //Rasterize.debug = ! Rasterize.debug;
             //RasterizeAntialias.debug = ! RasterizeAntialias.debug;
            }
            else if ('n' == c || 'N' == c)
            {
               // Move the near plane.
               if ('n' == c)
                  near += 0.01;
               else
                  near -= 0.01;
            }
            else if ('o' == c || 'O' == c)
            {
               // Change left, right, bottom, and top.
               if ('o' == c)
               {
                  left   += 0.1;
                  right  -= 0.1;
                  bottom += 0.1;
                  top    -= 0.1;
               }
               else
               {
                  left   -= 0.1;
                  right  += 0.1;
                  bottom -= 0.1;
                  top    += 0.1;
               }
            }
            else if (' ' == c)
            {
               render_z_buffer = ! render_z_buffer;
            }
            else if ('/' == c)
            {
               perspProjMatrix1Index = (perspProjMatrix1Index + 1) % 3;
            }
            else if ( Character.isDigit(c) )
            {
               int n = Integer.parseInt( Character.toString(c) );
               if (n < modelArray.size() - 2)
               {
                  scene.modelList.set(0, modelArray.get(n));
               }
            }
            else if ('a' == c)
            {
               Pipeline.doAntialiasing = ! Pipeline.doAntialiasing;
            }
            else if ('c' == c)
            {
               // Change the solid random color of each model.
               for (Model m : scene.modelList)
                     m.setColorRandom();
            }
            else if ('e' == c)
            {
               // Change the random color at each edge.
               for (Model m : scene.modelList)
                  for (LineSegment ls : m.lineSegmentList)
                     ls.setColorRandom();
            }
            else if ('p' == c)
            {
               scene.camera.perspective = ! scene.camera.perspective;
            }
            else if ('s' == c) // Scale the model 10% smaller.
            {
               scale *= 0.9;
            }
            else if ('S' == c) // Scale the model 10% larger.
            {
               scale *= 1.1;
            }
            else if ('x' == c)
            {
               xTranslation += 0.1;
            }
            else if ('X' == c)
            {
               xTranslation -= 0.1;
            }
            else if ('y' == c)
            {
               yTranslation += 0.1;
            }
            else if ('Y' == c)
            {
               yTranslation -= 0.1;
            }
            else if ('z' == c)
            {
               zTranslation += 0.1;
            }
            else if ('Z' == c)
            {
               zTranslation -= 0.1;
            }
            else if ('u' == c)
            {
               xRotation += 2.0;
            }
            else if ('U' == c)
            {
               xRotation -= 2.0;
            }
            else if ('v' == c)
            {
               yRotation += 2.0;
            }
            else if ('V' == c)
            {
               yRotation -= 2.0;
            }
            else if ('w' == c)
            {
               zRotation += 2.0;
            }
            else if ('W' == c)
            {
               zRotation -= 2.0;
            }

            // Set the model-to-view transformation matrix.
            // The order of the transformations is very important!
            Model model = scene.modelList.get(0);
            model.model2Identity();
            model.modelTranslate(xTranslation, yTranslation, zTranslation);
            model.modelRotate(xRotation, 1, 0, 0);
            model.modelRotate(yRotation, 0, 1, 0);
            model.modelRotate(zRotation, 0, 0, 1);
            model.modelScale(scale, scale, scale);

            // Set up the camera's view volume.
            if (scene.camera.perspective)
            {
               scene.camera.projPerspective(left, right, bottom, top, near);
               if (0 == perspProjMatrix1Index)
                  scene.camera.projMatrix = new PerspectiveProjectionMatrix1();
               else if (1 == perspProjMatrix1Index)
                  scene.camera.projMatrix = new PerspectiveProjectionMatrix2();
               else if (2 == perspProjMatrix1Index)
                  scene.camera.projMatrix = new PerspectiveProjectionMatrix3();
            }
            else
            {
               scene.camera.projOrtho(left, right, bottom, top);
            }
            if ('/'==c||'n'==c||'N'==c||'o'==c||'O'==c) System.out.print( scene.camera );

            // Render again.
            this.getFrameBuffer().clearFB(Color.black);
            Pipeline.render(scene, this.getFrameBuffer());
            if ( render_z_buffer )
            {
               FrameBuffer zBuf = this.getFrameBuffer().convertZB2FB();
               this.getFrameBuffer().setViewport(0, 0, zBuf);
            }
            repaint();
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
            //System.out.println( e );
            java.awt.Container c = (java.awt.Container)e.getSource();
            java.awt.Insets insets = c.getInsets();
            int w = c.getWidth()  - insets.left - insets.right;
            int h = c.getHeight() - insets.top  - insets.bottom;

            this.createFrameBuffer(w, h);
            this.getFrameBuffer().clearFB(Color.black);
            Pipeline.render(scene, this.getFrameBuffer());
            if ( render_z_buffer )
            {
               FrameBuffer zBuf = this.getFrameBuffer().convertZB2FB();
               this.getFrameBuffer().setViewport(0, 0, zBuf);
            }
            repaint();
         }
      };

      app.addKeyListener(app);
      app.addComponentListener(app);
   }
}
